import json
from typing import Dict, Any

import ray

from models.request import Request
from yodu import ESClient
from yodu.provider.provider_base import ProviderBase
from yodu.utils.utils import time_ago_to_date

meta = {
    "name": "TopItemsByUserAction",
    "version": "0.0.1",
    "author": "Shashank Agarwal",
    "description": "Return top items for a source by action_type \
        Example: Return top liked items for a given source \
        Example: Return top read items for a given source \
        Example: Return top read items for a given category",
    "args": {
        "action_type": {"type": "str", "source": "request.args.action_type"},
        "days_ago": {
            "type": "int",
            "source": "request.args.days_ago",
            "default": 1,
        },
        "user_id": {"type": "str", "source": "request.args.user_id"},
        "tag": {"type": "str", "source": "request.args.tag"},
        "next_token": {"type": "str", "source": "request.next_token"},
        "limit": {"type": "int", "source": "request.limit", "default": 10},
        "offset": {"type": "int", "source": "request.offset", "default": 0},
    },
}


@ray.remote
class Provider(ProviderBase):
    __es_client = None
    __indices = None
    __name = None

    def __init__(self, indices=Dict, name=str, **data: Any):
        super().__init__(**data)
        self.__name = name
        if not self.__es_client:
            self.__es_client = ESClient().get_client()
        self.__indices = indices

    def get_meta(self):
        return meta

    def build_input(self, config: Dict, request: Request):
        input_dict = {}
        """
        Get all values from request
        Then get all values from config
        Rest all values from meta
        """
        if request.limit:
            input_dict["limit"] = request.limit
        if request.offset:
            input_dict["offset"] = request.offset
        if request.user_id:
            input_dict["user_id"] = request.user_id
        if request.args:
            for key, val in request.args.items():
                input_dict[key] = val
        for name, val in config.items():
            if name not in input_dict:
                input_dict[name] = config[name]
        for name, val in meta["args"].items():
            if (
                    name not in input_dict
                    and name in meta["args"]
                    and "default" in meta["args"][name]
            ):
                input_dict[name] = meta["args"][name]["default"]
        return input_dict

    def execute(self, query):
        action_index = self.__indices["actions"]
        hits = self.__es_client.search(index=action_index, size=0, **query)
        res = hits.body["aggregations"]["top_tag_by_action"]["buckets"]
        """
        returns a list of top categories.
        We then get items from each of these categories and rank by count
        """
        items_dict = {}
        for hit in res:
            print(hit)
            for row in hit["top_action_hits"]["hits"]["hits"]:
                item = row["_source"]
                if "item_id" in item:
                    if item["item_id"] in items_dict:
                        items_dict[item["item_id"]] = (
                                items_dict[item["item_id"]] + 1
                        )
                    else:
                        items_dict[item["item_id"]] = 1
        items_dict = dict(
            sorted(items_dict.items(), key=lambda item: item[1], reverse=True)
        )
        top_items = list(items_dict.keys())
        return top_items

    def get_items(self, config: Dict, request: Request):
        """

        :param config:
        :param request:
        :return:
        """
        query = """
        {
          "query": {
                "bool": {
                    "must": {
                        "term": {
                            "type.keyword": "ACTION_TYPE_PLACEHOLDER"
                        }
                    },
                    "filter": {
                        "range": {
                            "created_at": {
                                "gt": "DATE_AGO_PLACEHOLDER",
                                "lt": "DATE_NOW_PLACEHOLDER"
                            }
                        }
                    }
                }
            },
          "aggs": {
            "top_tag_by_action": {
              "terms": {
                "field": "tags.TAG_PLACEHOLDER.keyword",
                "size": LIMIT_PLACEHOLDER
              },
              "aggs": {
                "top_action_hits": {
                  "top_hits": {
                    "sort": [
                      {
                        "created_at": {
                          "order": "desc"
                        }
                      }
                    ],
                    "_source": {
                      "includes": [ "item_id", "user_id","type" , "value", "tags" ]
                    },
                    "size": ITEMS_PER_ACTION
                  }
                }
              }
            }
          }
        }
        """
        input_dict = self.build_input(config=config, request=request)
        time_now = time_ago_to_date("now")
        time_ago = time_ago_to_date(str(input_dict["days_ago"]) + " days ago")
        query = query.replace("DATE_NOW_PLACEHOLDER", time_now)
        query = query.replace("DATE_AGO_PLACEHOLDER", time_ago)
        query = query.replace(
            "ACTION_TYPE_PLACEHOLDER", input_dict["action_type"]
        )
        query = query.replace("TAG_PLACEHOLDER", input_dict["tag"])
        query = query.replace("LIMIT_PLACEHOLDER", str(input_dict["limit"]))
        query = query.replace("ITEMS_PER_ACTION", str(5))
        res = json.loads(query)
        results = self.execute(res)
        return [self.__name, results]



import pprint

# initialize Steem class
from steem import Steem

"""
Download the entire blockchain and traverse all data
"""

s = Steem()

title = ""
# filters list
options = ["trending", "hot", "active", "created", "promoted"]
# get index and selected filter name
# option, index = pick(options, title)

query = {"limit": 2, "tag": ""}  # number of posts  # tag of posts
# #post list for selected query
# posts = {0: s.get_discussions_by_trending(query),
#          1: s.get_discussions_by_hot(query),
#          2: s.get_discussions_by_active(query),
#          3: s.get_discussions_by_created(query),
#          4: s.get_discussions_by_promoted(query)
# }

posts = s.get_discussions_by_created(query)
s.commit.post()

# print post list for selected filter
pprint.pprint(posts)



import uuid

from elasticsearch import helpers

from yodu import ESClient


class ItemHelper:
    es_client = None
    index_name = None

    def __init__(self, index_name, es_client: ESClient):
        assert es_client is not None
        assert index_name is not None
        self.es_client = es_client
        self.index_name = index_name

    def add(self, items: list):
        def convert_items(items_list):
            for item in items_list:
                if not item.id:
                    item.id = uuid.uuid4()
                doc = {
                    "_index": self.index_name,
                    "_id": str(item.id),
                    "_source": item.dict(),
                }
                yield doc

        helpers.bulk(self.es_client, convert_items(items))



from elasticsearch import Elasticsearch

from yodu import settings


class ESClient:
    client = None

    def __int__(self):
        pass

    def get_client(self):
        if not self.client:
            es_uri = (
                "https://"
                + settings.ES_USER
                + ":"
                + settings.ES_PASSWORD
                + "@"
                + settings.ES_HOST
                + ":"
                + str(settings.ES_PORT)
            )
            self.client = Elasticsearch(hosts=es_uri, verify_certs=False)
        return self.client

    def get_docs(self, item_index, item_ids: list):
        hits = self.get_client().mget(index=item_index, body={"ids": item_ids})
        return hits



from datetime import datetime
from typing import Optional, Dict
from uuid import uuid4

from pydantic import BaseModel, Field


class Item(BaseModel):
    # Unique ID for the item. If no ID is provided, a uuid4 is generated.
    id: str = Field(default_factory=uuid4)

    # Id of the source i.e. the publisher or the User who created this Item
    source: str

    # Properties dictionary to hold various metadata such as tags, categories, authors etc.
    props: Optional[Dict]

    # Internal identifier to specify if this item was provided by a Yodu Provider
    source_provider: Optional[str]

    # Time when the item was created
    created_at: datetime = Field(default_factory=datetime.now)

    # Time when the item was updated
    updated_at: datetime = Field(default_factory=datetime.now)



__title__ = "Yodu.ai Recommendation Engine"
__version__ = "0.0.3"
__author__ = "Shashank Agarwal"
__license__ = "Apache"
__copyright__ = "Copyright 2022-present The Geek Labs"

from db.es.es_client import ESClient
from recommeder.recommender import Recommender

TITLE = __title__
VERSION = __version__
AUTHOR = __author__
LICENSE = __license__
COPYRIGHT = __copyright__

DEFAULT_INDICES = ["items", "actions"]


def init():
    pass


def create_recommender(name: str):
    es = ESClient().get_client()
    indices = {}
    for index in DEFAULT_INDICES:
        indices[index] = name + "_" + index
    try:
        for key, val in indices.items():
            es.indices.create(index=val)
    except Exception as e:
        raise e
    return Recommender(name=name, indices=indices, es_client=es)


def get_recommender(name: str):
    es = ESClient().get_client()
    indices_res = es.indices.get(index=name + "_*")
    indices = {}

    if indices_res and indices_res.body:
        indices_dict = {}
        for index in indices_res.body:
            indices_dict[index] = 1
        for index in DEFAULT_INDICES:
            if name + "_" + index in indices_dict:
                indices[index] = name + "_" + index
    return Recommender(name=name, indices=indices, es_client=es)



from typing import Dict, Optional

from pydantic import BaseModel, Field


class Request(BaseModel):
    args: Optional[Dict]
    user_id: Optional[str]
    limit: Optional[int] = Field(default=10)
    offset: Optional[int] = Field(default=0)
    next_token: Optional[str]



import os

from dotenv import load_dotenv

load_dotenv()

ES_HOST = os.getenv("ES_HOST", "localhost")
ES_USER = os.getenv("ES_USER", "user")
ES_PASSWORD = os.getenv("ES_PASSWORD", "password")
ES_PORT = os.getenv("ES_PORT", 9200)
